/* microprotocols.c - minimalist and non-validating protocols implementation
 *
 * Copyright (C) 2003-2004 Federico Di Gregorio <fog@debian.org>
 *
 * This file is part of psycopg.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Python.h>
#include <structmember.h>

#define PSYCOPG_MODULE
#include "psycopg/config.h"
#include "psycopg/python.h"
#include "psycopg/psycopg.h"
#include "psycopg/cursor.h"
#include "psycopg/microprotocols.h"

/** the adapters registry **/

PyObject *psyco_adapters;

/* microprotocols_init - initialize the adapters dictionary */

int
microprotocols_init(PyObject *dict)
{
    /* create adapters dictionary and put it in module namespace */
    if ((psyco_adapters = PyDict_New()) == NULL) {
        return -1;
    }                         

    PyDict_SetItemString(dict, "adapters", psyco_adapters);
    
    return 0;
}


/* microprotocols_add - add a reverse type-caster to the dictionary */

int
microprotocols_add(PyTypeObject *type, PyObject *cast)
{
    Dprintf("microprotocols_add: cast %p for type '%s'", cast, type->tp_name);
    
    /* TODO: check the signature of "cast" and raise error if it is not the
       right one. */
    PyDict_SetItem(psyco_adapters, (PyObject*)type, cast);
    return 0;
}

/* microprotocols_adapt - adapt an object to the built-in protocol */

PyObject *
microprotocols_adapt(PyObject *obj, char *qattr)
{
    Dprintf("microprotocols_adapt: adapting %s (%s)",
            obj->ob_type->tp_name, qattr);
    
    /* if the object already has the method in 'qattr' we just Py_INCREF it
       and return (note that is is *not* checking protocol conformance, it is
       a usefull shortcut used by braindead microprotocols implementation. */
    if (PyObject_HasAttrString(obj, qattr)) {
        Dprintf("microprotocols_adapt:    -> same object");
        Py_INCREF(obj);
        return obj;
    }

    /* else look for an adapter in the registry */
    if (PyMapping_HasKey(psyco_adapters, (PyObject*)obj->ob_type)) {
        PyObject *adapter =
            PyDict_GetItem(psyco_adapters, (PyObject*)obj->ob_type);
        PyObject *adapted = PyObject_CallFunction(adapter, "O", obj);
        
        Dprintf("microprotocols_adapt:    -> %s",
                adapted ? adapted->ob_type->tp_name : NULL);

        return adapted;
    }

    /* else set the right exception and return NULL */
    Dprintf("microprotocols_adapt:    -> can't adapt");
    PyErr_SetString(ProgrammingError, "can't adapt argument");
    return NULL;
}

/** module-level functions **/

PyObject *
psyco_microprotocols_adapt(cursorObject *self, PyObject *args)
{
    PyObject *obj;
    char *qattr = MICROPROTOCOLS_GETQUOTED_NAME;

    if (!PyArg_ParseTuple(args, "O|s", &obj, &qattr)) return NULL;
    return microprotocols_adapt(obj, qattr);
}
